#include <errno.h>
#include <string.h>
#include <sys/fcntl.h>

#include <qregexp.h>
#include <kdebug.h>

#include "../config.h"


#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#ifdef HAVE_SYS_IOCTL_H
#include <sys/ioctl.h>
#endif

#define HAVE_PTY_H
#ifdef HAVE_PTY_H
#include <pty.h>
#endif

#ifdef HAVE_TERMIOS_H
#include <termios.h>
#endif

#ifdef HAVE_LIBUTIL_H
#include <libutil.h>
#endif


#include "fish.h"

// rich: hack to work for my system
#undef HAVE_GETPT
#define HAVE_OPENPTY

static int open_pty_pair(int fd[2])
{
#if defined(HAVE_GETPT) && defined(HAVE_TERMIOS_H)
/** with kind regards to The GNU C Library
Reference Manual for Version 2.2.x of the GNU C Library */
	int master, slave;
	char *name;
	struct ::termios ti;
	memset(&ti,0,sizeof(ti));

	ti.c_cflag = CLOCAL|CREAD|CS8;
	ti.c_cc[VMIN] = 1;

	master = getpt();
	if (master < 0) return 0;

	if (grantpt(master) < 0 || unlockpt(master) < 0) goto close_master;

	name = ptsname(master);
	if (name == NULL) goto close_master;

	slave = open(name, O_RDWR);
	if (slave == -1) goto close_master;

#if (defined(HAVE_ISASTREAM) || defined(isastream)) && defined(I_PUSH)
	if (isastream(slave) &&
		(ioctl(slave, I_PUSH, "ptem") < 0 ||
		 ioctl(slave, I_PUSH, "ldterm") < 0))
			goto close_slave;
#endif

	tcsetattr(slave, TCSANOW, &ti);
	fd[0] = master;
	fd[1] = slave;
	return 0;

close_slave:
	close(slave);

close_master:
	close(master);
	return -1;
#else
#ifdef HAVE_OPENPTY
	struct ::termios ti;
	memset(&ti,0,sizeof(ti));

	ti.c_cflag = CLOCAL|CREAD|CS8;
	ti.c_cc[VMIN] = 1;

	return openpty(fd,fd+1,NULL,&ti,NULL);
#else
#warning "No tty support available. Password dialog won't work."
	return socketpair(PF_UNIX,SOCK_STREAM,0,fd);
#endif
#endif
}

/**
creates the subprocess
*/
bool fishProtocol::connectionStart() {
	int fd[2];
	int rc, flags;
	thisFn = QString::null;

	rc = open_pty_pair(fd);
	if (rc == -1) {
		kdDebug() << "socketpair failed, error: " << strerror(errno) << endl;
		return true;
	}

	childPid = fork();
	if (childPid == -1) {
		kdDebug() << "fork failed, error: " << strerror(errno) << endl;
		close(fd[0]);
		close(fd[1]);
		return true;
	}
	if (childPid == 0) {
		// taken from konsole, see TEPty.C for details
		// note: if we're running on socket pairs,
		// this will fail, but thats what we expect

		for (int sig = 1; sig < NSIG; sig++) signal(sig,SIG_DFL);

		struct rlimit rlp;
		getrlimit(RLIMIT_NOFILE, &rlp);
		for (int i = 0; i < (int)rlp.rlim_cur; i++)
			if (i != fd[1]) close(i);

		dup2(fd[1],0);
		dup2(fd[1],1);
		dup2(fd[1],2);
		if (fd[1] > 2) close(fd[1]);

		setsid();

#if defined(TIOCSCTTY)
		ioctl(0, TIOCSCTTY, 0);
#endif

		int pgrp = getpid();
#ifdef _AIX
		tcsetpgrp(0, pgrp);
#else
		ioctl(0, TIOCSPGRP, (char *)&pgrp);
#endif

		const char *dev = ttyname(0);
		setpgid(0,0);
		if (dev) close(open(dev, O_WRONLY, 0));
		setpgid(0,0);


		QString ex("exec ssh -l '%u' -x -T -C -e none -p %p -q %h '%c'");
		ex.replace(QRegExp("%u"),connectionUser);
		ex.replace(QRegExp("%p"),QString::number(connectionPort));
		ex.replace(QRegExp("%h"),connectionHost);
		ex.replace(QRegExp("%c"),"echo FISH:;env TZ=UTC LANG=C LC_ALL=C LOCALE=C /bin/sh");
		execl("/bin/sh", "sh", "-c", ex.latin1(), NULL);
		kdDebug() << "could not exec " << ex << endl;
		exit(-1);
	}
	close(fd[1]);
	rc = fcntl(fd[0],F_GETFL,&flags);
	rc = fcntl(fd[0],F_SETFL,flags|O_NONBLOCK);
	childFd = fd[0];

	fd_set rfds, wfds;
	FD_ZERO(&rfds);
	FD_ZERO(&wfds);
	char buf[32768];
	int offset = 0;
	while (!isLoggedIn) {
		FD_SET(childFd,&rfds);
		FD_ZERO(&wfds);
		if (outBufPos >= 0) FD_SET(childFd,&wfds);
		rc = select(childFd+1, &rfds, &wfds, NULL, NULL);
		if (rc < 0) {
			kdDebug() << "select failed, rc: " << rc << ", error: " << strerror(errno) << endl;
			return true;
		}
		if (FD_ISSET(childFd,&wfds) && outBufPos >= 0) {
			if (outBuf > 0) rc = write(childFd,outBuf+outBufPos,outBufLen-outBufPos);
			else rc = 0;
			if (rc >= 0) outBufPos += rc;
			else {
				kdDebug() << "write failed, rc: " << rc << ", error: " << strerror(errno) << endl;
				outBufPos = -1;
				//return true;
			}
			if (outBufPos >= outBufLen) {
				outBufPos = -1;
				outBuf = NULL;
				outBufLen = 0;
			}
		}
		if (FD_ISSET(childFd,&rfds)) {
			rc = read(childFd,buf+offset,32768-offset);
			if (rc > 0) {
				int noff = establishConnection(buf,rc+offset);
				if (noff > 0) memmove(buf,buf+offset+rc-noff,noff);
				offset = noff;
			} else {
				kdDebug() << "read failed, rc: " << rc << ", error: " << strerror(errno) << endl;
				return true;
			}
		}
	}
	return false;
}





